function [W,V] = overdetsol(Z, R)

% [W,V] = overdetsol(Z,R)
%
% Time-stamp: </wrf/matlab/interp2/overdetsol.m, Wed, 30 Oct 2002, 19:13:29 EST, wrf@benvolio.ecse.rpi.edu>
%
% Do an overdetermined solution of the elevation points in the square
% matrix Z.  The unknown entries in Z marked by 0s.  Each known entry
% has an equation defining it to be its known value. All entries,
% known and unknown, are defined to be the average of their 4
% neighbors.  If an element doesn't have 4 neighbors, i.e., is on the
% border, then don't make an equation.  These latter equations are
% weighted by R relative to the former equations defining the known
% entries.  R very small will approximate a determined solution.  R
% very large will produce a surface of 0 divergence that ignores the
% known points.

% If R==0 then call detsol and don't refine.

% B is the output matrix of coefficients, and C the RHS vector.  W is
% the output matrix of elevations.

% Then, if REFINE is true, repeat the solution with the known points
% that had the highest errors being weighted higher.  Return that as
% V.

global B C W1

'new version'

if R==0
   W = detsol(Z);
   return
end

t= cputime;

REFINE=0

zl=min(min(Z(find(Z))));
zh=max(max(Z));
N=size(Z,1)
fprintf(1,'Input range, lo: %f, hi: %f\n',zl,zh)

N2=N*N;
N22 = (N-2)*(N-2);

[px,py] = find(Z);	% Find the indices of the known elts.
np=size(px,1)		% Number of nonzero elts.

C = zeros(N22, 1);       % RHS vector for the linear system.

NE=0;                   % Number of equations created so far. 

ns = 0;			% # entries in the list of sparse matrix B entries.
s = zeros(5*N22+np, 3); % The list of nonzero entries of B.


% Make an equation saying that each interior point is the average of its 4 
%  neighbors.

for i = 2:(N-1)
for j = 2:(N-1)
   NE=NE+1;
   ns = ns+1;
   s(ns,:) = [NE, N*j+i-N, 4*R];
   ns = ns+1;
   s(ns,:) = [NE, N*j+i-1-N, -R];
   ns = ns+1;
   s(ns,:) = [NE, N*j+i+1-N, -R]; 
   ns = ns+1;
   s(ns,:) = [NE, N*(j-1)+i-N, -R]; 
   ns = ns+1;
   s(ns,:) = [NE, N*(j+1)+i-N, -R]; 

   C(NE) = 0;
end
end

% Make an equation saying that each known point is equal to its value.

for k = 1:np
   NE=NE+1;
   i = px(k);
   j = py(k);
   ns = ns+1;
   s(ns,:) = [NE, N*j+i-N, 1];
   C(NE) = Z(i,j);
end


% Create the sparse coefficient matrix from the list of elements.

B = spconvert(s);

% whos

clear s                         % s is surprisingly large.

if ~REFINE
   clear px py
end

W1 = B \ C;			% Solve the overdetermined linear system.
W = reshape(W1,N,N);

% whos
clear W1

R

f=find(Z);
e = Z(f)-W(f);
fprintf(1, 'Average error: %f\n', mean(e))		% Should be practically zero.

e = abs(e);

global MEANABSERR MAXABSERR

MEANABSERR = mean(e);
fprintf(1, 'Average absolute error: %f or %f percent\n', MEANABSERR, MEANABSERR/(zh-zl)*100);

MAXABSERR = max(e);
fprintf(1, 'Max absolute error: %f or %f percent\n', MAXABSERR, MAXABSERR/(zh-zl)*100);

V=0;

if REFINE

% Re-solve with greater weights on the more inaccurate points.

weights = (1+10*((e/MAXABSERR).^2));

for k = 1:np
   i = px(k);
   j = py(k);
   B(N22+k,  N*j+i-N) = B(N22+k,  N*j+i-N) * weights(k);
end


np
size(weights)

C(N22+(1:np)) = C(N22+(1:np)) .* weights;

V1 = B \ C;			% Solve the overdetermined linear system.


V = reshape(V1,N,N);

% whos
clear V1

f=find(Z);
e = Z(f)-V(f);
fprintf(1, 'Average error: %f\n', mean(e))		% Should be practically zero.

e = abs(e);

global MEANABSERR2 MAXABSERR2

MEANABSERR2 = mean(e);
fprintf(1, 'Average absolute error after weighting: %f\n', MEANABSERR2)

MAXABSERR2 = max(e);
fprintf(1, 'Max abs error after weighting: %f\n', MAXABSERR2)

end

fprintf(1, 'CPU Time for overdetsol: %f\n', cputime-t)

